#include <stdio.h>
#include <stdlib.h>
#include <string.h>


typedef struct ensemble {
  unsigned int taille;
  int *tab;
} ensemble;

ensemble *allouer_ensemble(int taille) {
  ensemble *e;
  e = malloc(sizeof(struct ensemble));
  if(!e) perror(NULL);

  e->taille = taille;
  e->tab = calloc(taille, sizeof(int));
  return e;
}

void liberer_ensemble(ensemble *e) {
  free(e->tab);
  free(e);
}

int recherche_dichotomique(int *tab, int n, int debut, int fin) {
  int mediane;

  if (fin < debut)
    return 0;

  mediane = debut + ((fin - debut) / 2);

  if (tab[mediane] > n)
    return recherche_dichotomique(tab, n, debut, mediane-1);
  else if (tab[mediane] < n)
    return recherche_dichotomique(tab, n, mediane+1, fin);
  else
    return 1;
}


int membre_ensemble(int n, ensemble *e) {
  /* Solution naïve n'exploitant pas que le tableau est trié: */
  /* int i; */
  /* for(i=0; i<e->taille; i++) */
  /*   if (e->tab[i] == n)  */
  /*     return 1; */
  /* return 0; */

  /* Une solution plus efficace: */
  return recherche_dichotomique(e->tab, n, 0, (e->taille)-1);
}

ensemble *saisir_ensemble() {
  int i,t;
  ensemble *e;

  printf("taille? ");
  scanf("%d",&t);
  e = allouer_ensemble(t);

  for(i=0;i<t;i++) {
    printf("Élement %d? ",i);
    scanf("%d",&e->tab[i]);
  }

  return e;
}

void afficher_ensemble(ensemble *e) {
  int i,j=0;

  for (i=0; i<e->taille; i++)
    printf("%d ",e->tab[i]);
  printf("\n");

  for (i=0; i <= e->tab[e->taille-1]; i++) {
    if(membre_ensemble(i,e)) {
      printf("*");
    } else {
      printf(" ");
    }
  }
  printf("\n");
}

ensemble *union_ensembles(ensemble *e1, ensemble *e2) {
  ensemble *e;
  int i1=0,i2=0,i=0;
  int *tab;

  /* on alloue un tableau potentiellement trop grand */
  tab = calloc (sizeof(int),e1->taille + e2->taille);

  while (1) {
    if (i1 >= e1->taille && i2 >= e2->taille)
      break;
    if (i1 >= e1->taille)
      tab[i++] = e2->tab[i2++];
    else if (i2 >= e2->taille)
      tab[i++] = e1->tab[i1++];
    else if (e1->tab[i1] < e2->tab[i2])
      tab[i++] = e1->tab[i1++];
    else if (e1->tab[i1] > e2->tab[i2])
      tab[i++] = e2->tab[i2++];
    else { //(e1->tab[i1] == e2->tab[i2])
      tab[i++] = e2->tab[i2++];
      i1++;
    }
  }

  /* dont on copie la partie nécessaire vers un ensemble */
  e = allouer_ensemble(i);
  for(i1=0; i1<i; i1++)
    e->tab[i1] = tab[i1];

  free(tab);

  return e;
}


int main(void) {
  ensemble *e1, *e2, *e;
  e1 = saisir_ensemble();
  afficher_ensemble(e1);
  e2 = saisir_ensemble();
  afficher_ensemble(e2);
  printf("42 %s dans e1\n", membre_ensemble(42,e1) ? "est" : "n'est pas");

  e = union_ensembles(e1,e2);

  afficher_ensemble(e);

  liberer_ensemble(e1);
  liberer_ensemble(e2);
  liberer_ensemble(e);

  return 0;
}
